package com.sun.org.apache.xml.internal.dtm.ref;

import com.sun.org.apache.xml.internal.dtm.DTMException;
import com.sun.org.apache.xml.internal.res.XMLMessages;
import com.sun.org.apache.xml.internal.dtm.Axis;
import com.sun.org.apache.xml.internal.dtm.DTMAxisTraverser;
import com.sun.org.apache.xml.internal.utils.XMLStringFactory;
import com.sun.org.apache.xml.internal.dtm.DTMWSFilter;
import javax.xml.transform.Source;
import com.sun.org.apache.xml.internal.dtm.DTMManager;

public abstract class DTMDefaultBaseTraversers extends DTMDefaultBase
{
    public DTMDefaultBaseTraversers(DTMManager mgr, Source source, int dtmIdentity, DTMWSFilter whiteSpaceFilter, XMLStringFactory xstringfactory, boolean doIndexing) {
        super(mgr, source, dtmIdentity, whiteSpaceFilter, xstringfactory, doIndexing);
    }
    
    public DTMDefaultBaseTraversers(DTMManager mgr, Source source, int dtmIdentity, DTMWSFilter whiteSpaceFilter, XMLStringFactory xstringfactory, boolean doIndexing, int blocksize, boolean usePrevsib, boolean newNameTable) {
        super(mgr, source, dtmIdentity, whiteSpaceFilter, xstringfactory, doIndexing, blocksize, usePrevsib, newNameTable);
    }
    
    @Override
    public DTMAxisTraverser getAxisTraverser(int axis) {
        DTMAxisTraverser traverser;
        if (null == this.m_traversers) {
            this.m_traversers = new DTMAxisTraverser[Axis.getNamesLength()];
            traverser = null;
        }
        else {
            traverser = this.m_traversers[axis];
            if (traverser != null) {
                return traverser;
            }
        }
        switch (axis) {
            case 0: {
                traverser = new DTMDefaultBaseTraversers.AncestorTraverser();
                break;
            }
            case 1: {
                traverser = new DTMDefaultBaseTraversers.AncestorOrSelfTraverser();
                break;
            }
            case 2: {
                traverser = new DTMDefaultBaseTraversers.AttributeTraverser();
                break;
            }
            case 3: {
                traverser = new DTMDefaultBaseTraversers.ChildTraverser();
                break;
            }
            case 4: {
                traverser = new DTMDefaultBaseTraversers.DescendantTraverser();
                break;
            }
            case 5: {
                traverser = new DTMDefaultBaseTraversers.DescendantOrSelfTraverser();
                break;
            }
            case 6: {
                traverser = new DTMDefaultBaseTraversers.FollowingTraverser();
                break;
            }
            case 7: {
                traverser = new DTMDefaultBaseTraversers.FollowingSiblingTraverser();
                break;
            }
            case 9: {
                traverser = new DTMDefaultBaseTraversers.NamespaceTraverser();
                break;
            }
            case 8: {
                traverser = new DTMDefaultBaseTraversers.NamespaceDeclsTraverser();
                break;
            }
            case 10: {
                traverser = new DTMDefaultBaseTraversers.ParentTraverser();
                break;
            }
            case 11: {
                traverser = new DTMDefaultBaseTraversers.PrecedingTraverser();
                break;
            }
            case 12: {
                traverser = new DTMDefaultBaseTraversers.PrecedingSiblingTraverser();
                break;
            }
            case 13: {
                traverser = new DTMDefaultBaseTraversers.SelfTraverser();
                break;
            }
            case 16: {
                traverser = new DTMDefaultBaseTraversers.AllFromRootTraverser();
                break;
            }
            case 14: {
                traverser = new DTMDefaultBaseTraversers.AllFromNodeTraverser();
                break;
            }
            case 15: {
                traverser = new DTMDefaultBaseTraversers.PrecedingAndAncestorTraverser();
                break;
            }
            case 17: {
                traverser = new DTMDefaultBaseTraversers.DescendantFromRootTraverser();
                break;
            }
            case 18: {
                traverser = new DTMDefaultBaseTraversers.DescendantOrSelfFromRootTraverser();
                break;
            }
            case 19: {
                traverser = new DTMDefaultBaseTraversers.RootTraverser();
                break;
            }
            case 20: {
                return null;
            }
            default: {
                throw new DTMException(XMLMessages.createXMLMessage("ER_UNKNOWN_AXIS_TYPE", new Object[] { Integer.toString(axis) }));
            }
        }
        if (null == traverser) {
            throw new DTMException(XMLMessages.createXMLMessage("ER_AXIS_TRAVERSER_NOT_SUPPORTED", new Object[] { Axis.getNames(axis) }));
        }
        return this.m_traversers[axis] = traverser;
    }
    
    private class AncestorTraverser extends DTMAxisTraverser
    {
        @Override
        public int next(int context, int current) {
            return DTMDefaultBaseTraversers.this.getParent(current);
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current);
            while (-1 != (current = DTMDefaultBaseTraversers.this.m_parent.elementAt(current))) {
                if (DTMDefaultBaseTraversers.this.m_exptype.elementAt(current) != expandedTypeID) {
                    continue;
                }
                return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
            }
            return -1;
        }
    }
    
    private class AncestorOrSelfTraverser extends DTMDefaultBaseTraversers.AncestorTraverser
    {
        @Override
        public int first(int context) {
            return context;
        }
        
        @Override
        public int first(int context, int expandedTypeID) {
            return (DTMDefaultBaseTraversers.this.getExpandedTypeID(context) == expandedTypeID) ? context : this.next(context, context, expandedTypeID);
        }
    }
    
    private class AttributeTraverser extends DTMAxisTraverser
    {
        @Override
        public int next(int context, int current) {
            return (context == current) ? DTMDefaultBaseTraversers.this.getFirstAttribute(context) : DTMDefaultBaseTraversers.this.getNextAttribute(current);
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            current = ((context == current) ? DTMDefaultBaseTraversers.this.getFirstAttribute(context) : DTMDefaultBaseTraversers.this.getNextAttribute(current));
            while (DTMDefaultBaseTraversers.this.getExpandedTypeID(current) != expandedTypeID) {
                if (-1 != (current = DTMDefaultBaseTraversers.this.getNextAttribute(current))) {
                    continue;
                }
                return -1;
            }
            return current;
        }
    }
    
    private class ChildTraverser extends DTMAxisTraverser
    {
        protected int getNextIndexed(int axisRoot, int nextPotential, int expandedTypeID) {
            int nsIndex = DTMDefaultBaseTraversers.this.m_expandedNameTable.getNamespaceID(expandedTypeID);
            int lnIndex = DTMDefaultBaseTraversers.this.m_expandedNameTable.getLocalNameID(expandedTypeID);
            while (true) {
                int nextID = DTMDefaultBaseTraversers.this.findElementFromIndex(nsIndex, lnIndex, nextPotential);
                if (-2 != nextID) {
                    int parentID = DTMDefaultBaseTraversers.this.m_parent.elementAt(nextID);
                    if (parentID == axisRoot) {
                        return nextID;
                    }
                    if (parentID < axisRoot) {
                        return -1;
                    }
                    do {
                        parentID = DTMDefaultBaseTraversers.this.m_parent.elementAt(parentID);
                        if (parentID < axisRoot) {
                            return -1;
                        }
                        continue;
                    } while (parentID > axisRoot);
                    nextPotential = nextID + 1;
                }
                else {
                    DTMDefaultBaseTraversers.this.nextNode();
                    if (DTMDefaultBaseTraversers.this.m_nextsib.elementAt(axisRoot) != -2) {
                        return -1;
                    }
                    continue;
                }
            }
        }
        
        @Override
        public int first(int context) {
            return DTMDefaultBaseTraversers.this.getFirstChild(context);
        }
        
        @Override
        public int first(int context, int expandedTypeID) {
            int identity = DTMDefaultBaseTraversers.this.makeNodeIdentity(context);
            int firstMatch = this.getNextIndexed(identity, DTMDefaultBaseTraversers.this._firstch(identity), expandedTypeID);
            return DTMDefaultBaseTraversers.this.makeNodeHandle(firstMatch);
        }
        
        @Override
        public int next(int context, int current) {
            return DTMDefaultBaseTraversers.this.getNextSibling(current);
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            current = DTMDefaultBaseTraversers.this._nextsib(DTMDefaultBaseTraversers.this.makeNodeIdentity(current));
            while (-1 != current) {
                if (DTMDefaultBaseTraversers.this.m_exptype.elementAt(current) == expandedTypeID) {
                    return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
                }
                current = DTMDefaultBaseTraversers.this._nextsib(current);
            }
            return -1;
        }
    }
    
    private abstract class IndexedDTMAxisTraverser extends DTMAxisTraverser
    {
        protected final boolean isIndexed(int expandedTypeID) {
            return DTMDefaultBaseTraversers.this.m_indexing && 1 == DTMDefaultBaseTraversers.this.m_expandedNameTable.getType(expandedTypeID);
        }
        
        protected abstract boolean isAfterAxis(int p0, int p1);
        
        protected abstract boolean axisHasBeenProcessed(int p0);
        
        protected int getNextIndexed(int axisRoot, int nextPotential, int expandedTypeID) {
            int nsIndex = DTMDefaultBaseTraversers.this.m_expandedNameTable.getNamespaceID(expandedTypeID);
            int lnIndex = DTMDefaultBaseTraversers.this.m_expandedNameTable.getLocalNameID(expandedTypeID);
            while (true) {
                int next = DTMDefaultBaseTraversers.this.findElementFromIndex(nsIndex, lnIndex, nextPotential);
                if (-2 != next) {
                    if (this.isAfterAxis(axisRoot, next)) {
                        return -1;
                    }
                    return next;
                }
                else {
                    if (this.axisHasBeenProcessed(axisRoot)) {
                        return -1;
                    }
                    DTMDefaultBaseTraversers.this.nextNode();
                }
            }
        }
    }
    
    private class DescendantTraverser extends DTMDefaultBaseTraversers.IndexedDTMAxisTraverser
    {
        protected int getFirstPotential(int identity) {
            return identity + 1;
        }
        
        @Override
        protected boolean axisHasBeenProcessed(int axisRoot) {
            return DTMDefaultBaseTraversers.this.m_nextsib.elementAt(axisRoot) != -2;
        }
        
        protected int getSubtreeRoot(int handle) {
            return DTMDefaultBaseTraversers.this.makeNodeIdentity(handle);
        }
        
        protected boolean isDescendant(int subtreeRootIdentity, int identity) {
            return DTMDefaultBaseTraversers.this._parent(identity) >= subtreeRootIdentity;
        }
        
        @Override
        protected boolean isAfterAxis(int axisRoot, int identity) {
            while (identity != axisRoot) {
                identity = DTMDefaultBaseTraversers.this.m_parent.elementAt(identity);
                if (identity >= axisRoot) {
                    continue;
                }
                return true;
            }
            return false;
        }
        
        @Override
        public int first(int context, int expandedTypeID) {
            if (this.isIndexed(expandedTypeID)) {
                int identity = this.getSubtreeRoot(context);
                int firstPotential = this.getFirstPotential(identity);
                return DTMDefaultBaseTraversers.this.makeNodeHandle(this.getNextIndexed(identity, firstPotential, expandedTypeID));
            }
            return this.next(context, context, expandedTypeID);
        }
        
        @Override
        public int next(int context, int current) {
            int subtreeRootIdent = this.getSubtreeRoot(context);
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current) + 1;
            while (true) {
                int type = DTMDefaultBaseTraversers.this._type(current);
                if (!this.isDescendant(subtreeRootIdent, current)) {
                    return -1;
                }
                if (2 != type && 13 != type) {
                    return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
                }
                ++current;
            }
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            int subtreeRootIdent = this.getSubtreeRoot(context);
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current) + 1;
            if (this.isIndexed(expandedTypeID)) {
                return DTMDefaultBaseTraversers.this.makeNodeHandle(this.getNextIndexed(subtreeRootIdent, current, expandedTypeID));
            }
            while (true) {
                int exptype = DTMDefaultBaseTraversers.this._exptype(current);
                if (!this.isDescendant(subtreeRootIdent, current)) {
                    return -1;
                }
                if (exptype != expandedTypeID) {
                    ++current;
                }
                else {
                    return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
                }
            }
        }
    }
    
    private class DescendantOrSelfTraverser extends DTMDefaultBaseTraversers.DescendantTraverser
    {
        @Override
        protected int getFirstPotential(int identity) {
            return identity;
        }
        
        @Override
        public int first(int context) {
            return context;
        }
    }
    
    private class AllFromNodeTraverser extends DTMDefaultBaseTraversers.DescendantOrSelfTraverser
    {
        @Override
        public int next(int context, int current) {
            int subtreeRootIdent = DTMDefaultBaseTraversers.this.makeNodeIdentity(context);
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current) + 1;
            DTMDefaultBaseTraversers.this._exptype(current);
            if (!this.isDescendant(subtreeRootIdent, current)) {
                return -1;
            }
            return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
        }
    }
    
    private class FollowingTraverser extends DTMDefaultBaseTraversers.DescendantTraverser
    {
        @Override
        public int first(int context) {
            context = DTMDefaultBaseTraversers.this.makeNodeIdentity(context);
            int type = DTMDefaultBaseTraversers.this._type(context);
            if (2 == type || 13 == type) {
                context = DTMDefaultBaseTraversers.this._parent(context);
                int first = DTMDefaultBaseTraversers.this._firstch(context);
                if (-1 != first) {
                    return DTMDefaultBaseTraversers.this.makeNodeHandle(first);
                }
            }
            int first;
            do {
                first = DTMDefaultBaseTraversers.this._nextsib(context);
                if (-1 == first) {
                    context = DTMDefaultBaseTraversers.this._parent(context);
                }
                else {
                    continue;
                }
            } while (-1 == first && -1 != context);
            return DTMDefaultBaseTraversers.this.makeNodeHandle(first);
        }
        
        @Override
        public int first(int context, int expandedTypeID) {
            int type = DTMDefaultBaseTraversers.this.getNodeType(context);
            if (2 == type || 13 == type) {
                context = DTMDefaultBaseTraversers.this.getParent(context);
                int first = DTMDefaultBaseTraversers.this.getFirstChild(context);
                if (-1 != first) {
                    if (DTMDefaultBaseTraversers.this.getExpandedTypeID(first) == expandedTypeID) {
                        return first;
                    }
                    return this.next(context, first, expandedTypeID);
                }
            }
            int first;
            do {
                first = DTMDefaultBaseTraversers.this.getNextSibling(context);
                if (-1 == first) {
                    context = DTMDefaultBaseTraversers.this.getParent(context);
                }
                else {
                    if (DTMDefaultBaseTraversers.this.getExpandedTypeID(first) == expandedTypeID) {
                        return first;
                    }
                    return this.next(context, first, expandedTypeID);
                }
            } while (-1 == first && -1 != context);
            return first;
        }
        
        @Override
        public int next(int context, int current) {
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current);
            while (true) {
                ++current;
                int type = DTMDefaultBaseTraversers.this._type(current);
                if (-1 == type) {
                    return -1;
                }
                if (2 == type) {
                    continue;
                }
                if (13 == type) {
                    continue;
                }
                return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
            }
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current);
            while (true) {
                ++current;
                int etype = DTMDefaultBaseTraversers.this._exptype(current);
                if (-1 == etype) {
                    return -1;
                }
                if (etype != expandedTypeID) {
                    continue;
                }
                return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
            }
        }
    }
    
    private class FollowingSiblingTraverser extends DTMAxisTraverser
    {
        @Override
        public int next(int context, int current) {
            return DTMDefaultBaseTraversers.this.getNextSibling(current);
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            while (-1 != (current = DTMDefaultBaseTraversers.this.getNextSibling(current))) {
                if (DTMDefaultBaseTraversers.this.getExpandedTypeID(current) != expandedTypeID) {
                    continue;
                }
                return current;
            }
            return -1;
        }
    }
    
    private class NamespaceDeclsTraverser extends DTMAxisTraverser
    {
        @Override
        public int next(int context, int current) {
            return (context == current) ? DTMDefaultBaseTraversers.this.getFirstNamespaceNode(context, false) : DTMDefaultBaseTraversers.this.getNextNamespaceNode(context, current, false);
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            current = ((context == current) ? DTMDefaultBaseTraversers.this.getFirstNamespaceNode(context, false) : DTMDefaultBaseTraversers.this.getNextNamespaceNode(context, current, false));
            while (DTMDefaultBaseTraversers.this.getExpandedTypeID(current) != expandedTypeID) {
                if (-1 != (current = DTMDefaultBaseTraversers.this.getNextNamespaceNode(context, current, false))) {
                    continue;
                }
                return -1;
            }
            return current;
        }
    }
    
    private class NamespaceTraverser extends DTMAxisTraverser
    {
        @Override
        public int next(int context, int current) {
            return (context == current) ? DTMDefaultBaseTraversers.this.getFirstNamespaceNode(context, true) : DTMDefaultBaseTraversers.this.getNextNamespaceNode(context, current, true);
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            current = ((context == current) ? DTMDefaultBaseTraversers.this.getFirstNamespaceNode(context, true) : DTMDefaultBaseTraversers.this.getNextNamespaceNode(context, current, true));
            while (DTMDefaultBaseTraversers.this.getExpandedTypeID(current) != expandedTypeID) {
                if (-1 != (current = DTMDefaultBaseTraversers.this.getNextNamespaceNode(context, current, true))) {
                    continue;
                }
                return -1;
            }
            return current;
        }
    }
    
    private class ParentTraverser extends DTMAxisTraverser
    {
        @Override
        public int first(int context) {
            return DTMDefaultBaseTraversers.this.getParent(context);
        }
        
        @Override
        public int first(int current, int expandedTypeID) {
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current);
            while (-1 != (current = DTMDefaultBaseTraversers.this.m_parent.elementAt(current))) {
                if (DTMDefaultBaseTraversers.this.m_exptype.elementAt(current) != expandedTypeID) {
                    continue;
                }
                return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
            }
            return -1;
        }
        
        @Override
        public int next(int context, int current) {
            return -1;
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            return -1;
        }
    }
    
    private class PrecedingTraverser extends DTMAxisTraverser
    {
        protected boolean isAncestor(int contextIdent, int currentIdent) {
            contextIdent = DTMDefaultBaseTraversers.this.m_parent.elementAt(contextIdent);
            while (-1 != contextIdent) {
                if (contextIdent == currentIdent) {
                    return true;
                }
                contextIdent = DTMDefaultBaseTraversers.this.m_parent.elementAt(contextIdent);
            }
            return false;
        }
        
        @Override
        public int next(int context, int current) {
            int subtreeRootIdent = DTMDefaultBaseTraversers.this.makeNodeIdentity(context);
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current) - 1;
            while (current >= 0) {
                short type = DTMDefaultBaseTraversers.this._type(current);
                if (2 != type && 13 != type && !this.isAncestor(subtreeRootIdent, current)) {
                    return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
                }
                --current;
            }
            return -1;
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            int subtreeRootIdent = DTMDefaultBaseTraversers.this.makeNodeIdentity(context);
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current) - 1;
            while (current >= 0) {
                int exptype = DTMDefaultBaseTraversers.this.m_exptype.elementAt(current);
                if (exptype == expandedTypeID && !this.isAncestor(subtreeRootIdent, current)) {
                    return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
                }
                --current;
            }
            return -1;
        }
    }
    
    private class PrecedingAndAncestorTraverser extends DTMAxisTraverser
    {
        @Override
        public int next(int context, int current) {
            int subtreeRootIdent = DTMDefaultBaseTraversers.this.makeNodeIdentity(context);
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current) - 1;
            while (current >= 0) {
                short type = DTMDefaultBaseTraversers.this._type(current);
                if (2 != type && 13 != type) {
                    return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
                }
                --current;
            }
            return -1;
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            int subtreeRootIdent = DTMDefaultBaseTraversers.this.makeNodeIdentity(context);
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current) - 1;
            while (current >= 0) {
                int exptype = DTMDefaultBaseTraversers.this.m_exptype.elementAt(current);
                if (exptype != expandedTypeID) {
                    --current;
                }
                else {
                    return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
                }
            }
            return -1;
        }
    }
    
    private class PrecedingSiblingTraverser extends DTMAxisTraverser
    {
        @Override
        public int next(int context, int current) {
            return DTMDefaultBaseTraversers.this.getPreviousSibling(current);
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            while (-1 != (current = DTMDefaultBaseTraversers.this.getPreviousSibling(current))) {
                if (DTMDefaultBaseTraversers.this.getExpandedTypeID(current) != expandedTypeID) {
                    continue;
                }
                return current;
            }
            return -1;
        }
    }
    
    private class SelfTraverser extends DTMAxisTraverser
    {
        @Override
        public int first(int context) {
            return context;
        }
        
        @Override
        public int first(int context, int expandedTypeID) {
            return (DTMDefaultBaseTraversers.this.getExpandedTypeID(context) == expandedTypeID) ? context : -1;
        }
        
        @Override
        public int next(int context, int current) {
            return -1;
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            return -1;
        }
    }
    
    private class AllFromRootTraverser extends DTMDefaultBaseTraversers.AllFromNodeTraverser
    {
        @Override
        public int first(int context) {
            return DTMDefaultBaseTraversers.this.getDocumentRoot(context);
        }
        
        @Override
        public int first(int context, int expandedTypeID) {
            return (DTMDefaultBaseTraversers.this.getExpandedTypeID(DTMDefaultBaseTraversers.this.getDocumentRoot(context)) == expandedTypeID) ? context : this.next(context, context, expandedTypeID);
        }
        
        @Override
        public int next(int context, int current) {
            int subtreeRootIdent = DTMDefaultBaseTraversers.this.makeNodeIdentity(context);
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current) + 1;
            int type = DTMDefaultBaseTraversers.this._type(current);
            if (type == -1) {
                return -1;
            }
            return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            int subtreeRootIdent = DTMDefaultBaseTraversers.this.makeNodeIdentity(context);
            current = DTMDefaultBaseTraversers.this.makeNodeIdentity(current) + 1;
            while (true) {
                int exptype = DTMDefaultBaseTraversers.this._exptype(current);
                if (exptype == -1) {
                    return -1;
                }
                if (exptype != expandedTypeID) {
                    ++current;
                }
                else {
                    return DTMDefaultBaseTraversers.this.makeNodeHandle(current);
                }
            }
        }
    }
    
    private class RootTraverser extends DTMDefaultBaseTraversers.AllFromRootTraverser
    {
        @Override
        public int first(int context, int expandedTypeID) {
            int root = DTMDefaultBaseTraversers.this.getDocumentRoot(context);
            return (DTMDefaultBaseTraversers.this.getExpandedTypeID(root) == expandedTypeID) ? root : -1;
        }
        
        @Override
        public int next(int context, int current) {
            return -1;
        }
        
        @Override
        public int next(int context, int current, int expandedTypeID) {
            return -1;
        }
    }
    
    private class DescendantOrSelfFromRootTraverser extends DTMDefaultBaseTraversers.DescendantTraverser
    {
        @Override
        protected int getFirstPotential(int identity) {
            return identity;
        }
        
        @Override
        protected int getSubtreeRoot(int handle) {
            return DTMDefaultBaseTraversers.this.makeNodeIdentity(DTMDefaultBaseTraversers.this.getDocument());
        }
        
        @Override
        public int first(int context) {
            return DTMDefaultBaseTraversers.this.getDocumentRoot(context);
        }
        
        @Override
        public int first(int context, int expandedTypeID) {
            if (this.isIndexed(expandedTypeID)) {
                int identity = 0;
                int firstPotential = this.getFirstPotential(identity);
                return DTMDefaultBaseTraversers.this.makeNodeHandle(this.getNextIndexed(identity, firstPotential, expandedTypeID));
            }
            int root = this.first(context);
            return this.next(root, root, expandedTypeID);
        }
    }
    
    private class DescendantFromRootTraverser extends DTMDefaultBaseTraversers.DescendantTraverser
    {
        @Override
        protected int getFirstPotential(int identity) {
            return DTMDefaultBaseTraversers.this._firstch(0);
        }
        
        @Override
        protected int getSubtreeRoot(int handle) {
            return 0;
        }
        
        @Override
        public int first(int context) {
            return DTMDefaultBaseTraversers.this.makeNodeHandle(DTMDefaultBaseTraversers.this._firstch(0));
        }
        
        @Override
        public int first(int context, int expandedTypeID) {
            if (this.isIndexed(expandedTypeID)) {
                int identity = 0;
                int firstPotential = this.getFirstPotential(identity);
                return DTMDefaultBaseTraversers.this.makeNodeHandle(this.getNextIndexed(identity, firstPotential, expandedTypeID));
            }
            int root = DTMDefaultBaseTraversers.this.getDocumentRoot(context);
            return this.next(root, root, expandedTypeID);
        }
    }
}
